/// \file ijkgrid.txx
/// ijk templates defining regular grid classes
/// Version 0.1.0

/*
  IJK: Isosurface Jeneration Kode
  Copyright (C) 2008 Rephael Wenger

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _IJKGRID_
#define _IJKGRID_

namespace IJKGRID {

  // **************************************************
  // TYPE DEFINITIONS
  // **************************************************

  typedef long GRID_SIZE_TYPE;

  // **************************************************
  // TEMPLATE CLASS GRID
  // **************************************************

  template <class DTYPE, class ATYPE, class VTYPE> class GRID {

  protected:
    DTYPE dimension;     // grid dimension
    ATYPE * axis_size;   // axis_size[i] = # grid points along axis i
    GRID_SIZE_TYPE num_vertices;  // number of grid vertices

    void Init(const DTYPE dimension, const ATYPE * axis_size);
    void Set(const DTYPE dimension, const ATYPE * axis_size);
    void FreeAll();

  public:
    GRID(const DTYPE dimension, const ATYPE * axis_size);
    ~GRID();
    GRID(const GRID & grid);                      // copy constructor
    const GRID & operator = (const GRID & right); // copy assignment

    // get functions
    DTYPE Dimension() const { return(dimension); }
    const ATYPE * AxisSize() const { return(axis_size); }
    ATYPE AxisSize(const DTYPE i) const { return(axis_size[i]); }
    GRID_SIZE_TYPE NumVertices() const { return(num_vertices); }

    // compute functions
    GRID_SIZE_TYPE ComputeNumVertices
    (const VTYPE iv0, const VTYPE iv1) const;
    GRID_SIZE_TYPE ComputeNumCubes() const;

    // check function
    bool Check(const DTYPE dimension, const ATYPE * axis_size,
	       IJK::ERROR & error);
  };

  // **************************************************
  // TEMPLATE CLASS SCALAR_GRID
  // **************************************************

  // class storing scalar grid information
  template <class DTYPE, class ATYPE, class VTYPE, class STYPE>
  class SCALAR_GRID:public GRID<DTYPE,ATYPE,VTYPE> {

  protected:
    STYPE * scalar;
    // point (x0,x1,x2,...) has scalar value
    //     scalar_grid[x0 + x1*axis_size[0] + 
    //                   x2*axis_size[0]*axis_size[1] + ...]

  public:
    SCALAR_GRID(const DTYPE dimension, const ATYPE * axis_size,
		STYPE * scalar);
    ~SCALAR_GRID(){};
    // Note: constructor and destructor do not allocate or free scalar memory

    // set functions
    template <class GTYPE>
    void Replace(const SCALAR_GRID<DTYPE,ATYPE,VTYPE,STYPE> & scalar_grid2,
		 const VTYPE iv0, const VTYPE iv1);

    // get functions
    const STYPE * Scalar() const { return(scalar); };
  };

  // **************************************************
  // TEMPLATE FUNCTIONS: COUNTING AND INDEXING
  // **************************************************

  /// Return coordinate of specified vertex
  template <class VTYPE, class DTYPE, class ATYPE, class GTYPE>
  void compute_coord(const VTYPE iv, const DTYPE dimension,
		     const ATYPE * axis_size, GTYPE * coord)
  // compute coordinates of grid vertex iv
  // Precondition: axis_size[d] > 0 for all d = 0,...,dimension-1
  {
    VTYPE k = iv;
    for (DTYPE d = 0; d < dimension; d++) {
      coord[d] = k % axis_size[d];
      k = k / axis_size[d];
    };
  }

  /// Return index of vertex with specified coord
  template <class VTYPE, class GTYPE, class DTYPE, class ATYPE>
  VTYPE compute_vertex_index
  (const GTYPE * coord, const DTYPE dimension, const ATYPE * axis_size)
  {
    VTYPE iv = 0;
    VTYPE inc = 1;
    for (DTYPE d = 0; d < dimension; d++) {
      iv += inc*coord[d];
      inc = inc*axis_size[d];
    }

    return(iv);
  }

  /// Return number of grid vertices
  template <class DTYPE, class ATYPE>
  GRID_SIZE_TYPE compute_num_grid_vertices
  (const DTYPE dimension, const ATYPE * axis_size)
  {
    GRID_SIZE_TYPE num_vertices = 1;
    for (DTYPE d = 0; d < dimension; d++)
      { num_vertices = num_vertices * axis_size[d]; }
    return(num_vertices);
  }

  /// Return number of grid vertices between two vertices
  template <class VTYPE, class DTYPE, class ATYPE>
  GRID_SIZE_TYPE compute_num_grid_vertices
  (const DTYPE dimension, const ATYPE * axis_size,
   const VTYPE iv0, const VTYPE iv1)
  // get number of grid vertices between iv0 and iv1
  // Precondition: 0 <= iv0 <= iv1 < total_num_grid_vertices
  {
    VTYPE coord0[dimension];
    VTYPE coord1[dimension];
    IJK::PROCEDURE_ERROR error("compute_num_grid_vertices");

    if (!check_range(dimension, axis_size, iv0, iv1, error)) { throw error; };

    compute_coord(iv0, dimension, axis_size, coord0);
    compute_coord(iv1, dimension, axis_size, coord1);

    GRID_SIZE_TYPE num_grid_vertices = 1;
    for (DTYPE d = 0; d < dimension; d++) {
      if (coord0[d] > coord1[d]) {
	error.AddMessage("Programming error in calculating ", d,
			 "'th coordinate.");
	error.AddMessage("  coord0 (", coord0[d], 
			 ") > coord1 (", coord1[d], ").");
	throw error;
      }

      num_grid_vertices = num_grid_vertices * (coord1[d]-coord0[d]+1);
    }

    return(num_grid_vertices);
  }

  /// Return number of grid cubes
  template <class DTYPE, class ATYPE>
  GRID_SIZE_TYPE compute_num_grid_cubes
  (const DTYPE dimension, const ATYPE * axis_size)
  {
    GRID_SIZE_TYPE num_cubes = 1;
    for (DTYPE d = 0; d < dimension; d++) {
      if (axis_size[d] < 2) { num_cubes = 0; }
      else { num_cubes = num_cubes * (axis_size[d]-1); }
    }
    return(num_cubes);
  }

  /// Return number of cube vertices
  template <class DTYPE> 
  long compute_num_cube_vertices(const DTYPE dimension)
  { return(1L << dimension); };

  /// Return number of cube facet vertices
  template <class DTYPE> 
  long compute_num_cube_facet_vertices(const DTYPE dimension)
  // Precondition: dimension > 0
  { return(1L << dimension-1); };

  /// Check range of vertices
  template <class VTYPE, class DTYPE, class ATYPE>
  bool check_range
  (const DTYPE dimension, const ATYPE * axis_size,
   const VTYPE iv0, const VTYPE iv1, IJK::ERROR & error)
  // return true if 0 <= iv0 <= iv1 < total_num_grid_vertices
  {
    const VTYPE total_num_grid_vertices =
      compute_num_grid_vertices<VTYPE>(dimension, axis_size);

    if (iv0 > iv1) {
      error.AddMessage("Illegal vertex range. Vertex index ", iv0, 
		       " is greater than vertex index ", iv1, ".");
      return(false);
    }

    if (0 > iv0 || iv1 >= total_num_grid_vertices) {
      error.AddMessage("Illegal vertex indices: ", iv0, ", ", iv1, ".");
      error.AddMessage("Vertex indices should be in range: [0,",
		       total_num_grid_vertices, "].");
      return(false);
    }

    return(true);
  }

  // **************************************************
  // TEMPLATE FUNCTIONS: GETTING VERTICES
  // **************************************************

  // WARNING: Procedure get_grid_vertices() has not been tested.
  template <class DTYPE, class ATYPE, class VTYPE>
  void get_grid_vertices
  (const DTYPE dimension, const ATYPE * axis_size, 
   const VTYPE iv0, const VTYPE iv1, VTYPE * vlist)
  // get list of grid vertices between iv0 and iv1
  // Precondition: 0 <= iv0 <= iv1 < total_num_grid_vertices
  // Precondition: vlist[] is preallocated to size at least num_grid_vertices
  {
    VTYPE coord0[dimension];
    VTYPE coord1[dimension];
    VTYPE coord[dimension];
    VTYPE axis_increment[dimension];
    IJK::PROCEDURE_ERROR error("get_grid_vertices");

    if (dimension <= 0) { return; };

    if (axis_size == NULL) {
      error.AddMessage("Programming error. NULL axis size.");
      throw error;
    }

    if (!check_range(dimension, axis_size, iv0, iv1, error)) { throw error; };

    VTYPE num_grid_vertices = 
      compute_num_grid_vertices(dimension, axis_size, iv0, iv1);

    if (num_grid_vertices < 1) return;

    if (vlist == NULL) {
      error.AddMessage("Programming error. NULL vlist.");
      throw error;
    }

    compute_increment(dimension, axis_size, axis_increment);
    compute_coord(iv0, dimension, axis_size, coord0);
    compute_coord(iv1, dimension, axis_size, coord1);

    // initialize prev_num_cubes
    vlist[0] = iv0;
    VTYPE prev_num_vert = 1;

    // Process axes 0,1,2,..., dimension-1
    VTYPE * vcur_ptr = vlist + prev_num_vert;
    for (DTYPE d = 0; d < dimension; d++) {
      VTYPE v_inc = axis_increment[d];
      for (VTYPE c = coord0[d]+1; c <= coord1[d]; c++) {
	for (VTYPE * vprev_ptr = vlist; 
	     vprev_ptr != vlist+prev_num_vert; vprev_ptr++) {
	  *(vcur_ptr) = v_inc + *(vprev_ptr);
	  vcur_ptr++;
	};
	v_inc = v_inc + axis_increment[d];
      }
      prev_num_vert = prev_num_vert*(coord1[d]-coord0[d]+1);
    }

    if (prev_num_vert != num_grid_vertices) {
      error.AddMessage("Programming error.  Added ", prev_num_vert, 
		       " grid vertices to list.");
      error.AddMessage("Number of grid vertices between vertex ", iv0,
		       " and vertex ", iv1, " is ", num_grid_vertices, ".");
      throw error;
    }
  }

  // WARNING: Procedure get_primary_cube_vertices() has not been tested.
  template <class DTYPE, class ATYPE, class VTYPE>
  void get_primary_cube_vertices
  (const DTYPE dimension, const ATYPE * axis_size, VTYPE * vlist)
  // get list of primary cube vertices (lowest coord vertices in cubes)
  // vlist[k] = list of primary cube vertices
  //    Precondition: vlist[] is preallocated to length at least 
  //                  number of cubes
  {
    IJK::PROCEDURE_ERROR error("get_primary_cube_vertices");
    VTYPE axis_increment[dimension];

    if (dimension <= 0) { return; };

    if (axis_size == NULL) {
      error.AddMessage("Programming error. NULL axis size.");
      throw error;
    }

    const VTYPE num_cubes = 
      compute_num_grid_cubes(dimension, axis_size);
    if (num_cubes < 1) return;
    // Note: axis_size[d] >= 1 for all d. dimension >= 1.

    if (vlist == NULL) {
      error.AddMessage("Programming error. NULL vlist.");
      throw error;
    }

    compute_increment(dimension, axis_size, axis_increment);

    // initialize prev_num_cubes
    VTYPE prev_num_cubes = 0;

    // Process first axis separately for faster execution time
    // (Makes little difference unless first axis is very long)
    if (dimension > 0) {
      if (axis_size[0] < 1) {
	error.AddMessage("Programming error. axis_size[0] < 1.");
	throw error;
      }
      if (axis_increment[0] != 1) {
	error.AddMessage("Programming error. axis_increment[0] != 1.");
	throw error;
      }

      for (VTYPE i = 0; i < axis_size[0]-1; i++) {
	vlist[i] = i;
      }
      prev_num_cubes = axis_size[0]-1;
    }

    // Process axes 1,2,..., dimension-1
    VTYPE * vcur_ptr = vlist + prev_num_cubes;
    for (int d = 1; d < dimension; d++) {
      if (axis_increment[d] < 1) {
	error.AddMessage("Programming error. axis_increment[", d,
			 "] < 1.");
	throw error;
      }

      VTYPE iv0 = axis_increment[d];
      for (VTYPE i = 1; i < axis_size[d]-1; i++) {
	for (VTYPE * vprev_ptr = vlist; 
	     vprev_ptr != vlist+prev_num_cubes; vprev_ptr++) {
	  *(vcur_ptr) = iv0 + *(vprev_ptr);
	  vcur_ptr++;
	};
	iv0 = iv0 + axis_increment[d];
      }
      prev_num_cubes = prev_num_cubes*(axis_size[d]-1);
    }

    if (prev_num_cubes != num_cubes) {
      error.AddMessage("Programming error.  Added ", prev_num_cubes, 
		       " primary cube vertices to list.");
      error.AddMessage("Number of primary cube vertices should be ", 
		       num_cubes, ".");
      throw error;
    }
  }

  // **************************************************
  // TEMPLATE CLASS GRID MEMBER FUNCTIONS
  // **************************************************

  template <class DTYPE, class ATYPE, class VTYPE> 
  GRID<DTYPE,ATYPE,VTYPE>::GRID<DTYPE,ATYPE,VTYPE>
  (const DTYPE dimension, const ATYPE * axis_size)
  // constructor
  {
    Init(dimension, axis_size);
  }

  template <class DTYPE, class ATYPE, class VTYPE> 
  GRID<DTYPE,ATYPE,VTYPE>::~GRID<DTYPE,ATYPE,VTYPE>()
  // destructor
  {
    FreeAll();
  }

  template <class DTYPE, class ATYPE, class VTYPE> 
  void GRID<DTYPE,ATYPE,VTYPE>::Init
  (const DTYPE dimension, const ATYPE * axis_size)
  {
    this->axis_size = NULL;
    this->dimension = 0;
    this->num_vertices = 0;
    Set(dimension, axis_size);
  }

  template <class DTYPE, class ATYPE, class VTYPE> 
  void GRID<DTYPE,ATYPE,VTYPE>::Set
  (const DTYPE dimension, const ATYPE * axis_size)
  {
    FreeAll();
    this->dimension = dimension;
    this->axis_size = new ATYPE[dimension];
    for (DTYPE d = 0; d < dimension; d++)
      { this->axis_size[d] = axis_size[d]; }
    this->num_vertices = 
      compute_num_grid_vertices<VTYPE>(dimension, axis_size);
  }

  template <class DTYPE, class ATYPE, class VTYPE> 
  void GRID<DTYPE,ATYPE,VTYPE>::FreeAll()
  {
    if (axis_size == NULL) { delete [] axis_size; };
    axis_size = NULL;
    dimension = 0;
    num_vertices = 0;
  }

  template <class DTYPE, class ATYPE, class VTYPE> 
  GRID<DTYPE,ATYPE,VTYPE>::
  GRID<DTYPE,ATYPE,VTYPE>(const GRID<DTYPE,ATYPE,VTYPE> & grid)
  // copy constructor
  {
    Init(grid.Dimension(), grid.AxisSize());
  }

  template <class DTYPE, class ATYPE, class VTYPE> 
  const GRID<DTYPE,ATYPE,VTYPE> & 
  GRID<DTYPE,ATYPE,VTYPE>::operator = (const GRID<DTYPE,ATYPE,VTYPE> & right)
  // copy assignment	
  {
    if (&right != this) {         // avoid self-assignment
      Set(right.Dimension(), right.AxisSize());
    }
  }

  template <class DTYPE, class ATYPE, class VTYPE>
  GRID_SIZE_TYPE GRID<DTYPE,ATYPE,VTYPE>::
  ComputeNumVertices(const VTYPE iv0, const VTYPE iv1) const
  {
    return(compute_num_vertices(Dimension(), AxisSize(), iv0, iv1));
  }

  template <class DTYPE, class ATYPE, class VTYPE>
  GRID_SIZE_TYPE GRID<DTYPE,ATYPE,VTYPE>::ComputeNumCubes() const
  {
    return(compute_num_grid_cubes(Dimension(), AxisSize()));
  }

  template <class DTYPE, class ATYPE, class VTYPE> 
  bool GRID<DTYPE,ATYPE,VTYPE>::Check
  (const DTYPE dimension, const ATYPE * axis_size, IJK::ERROR & error)
  // return true if grid dimension and axis size match parameter dimension
  //      and parameter axis size, respectively
  {
    if (dimension != this->dimension) {
      error.AddMessage("Incorrect grid dimension ", this->dimension, ".");
      return(false);
    }

    for (int d = 0; d < dimension; d++) {
      if (axis_size[d] != this->axis_size[d]) {
	error.AddMessage("Illegal axis size[", d, "] = ", 
			 this->axis_size[d], ".");
	return(false);
      }
    }

    VTYPE num_vertices = 
      compute_num_grid_vertices<VTYPE>(dimension, axis_size);

    if (num_vertices != this->num_vertices) {
      error.AddMessage("Incorrect number of grid vertices ", 
		       this->num_vertices, ".");
      return(false);
    }

    return(true);
  }

  // **************************************************
  // TEMPLATE CLASS SCALAR_GRID MEMBER FUNCTIONS
  // **************************************************

  template <class DTYPE, class ATYPE, class VTYPE, class STYPE>
  SCALAR_GRID<DTYPE,ATYPE,VTYPE,STYPE>::SCALAR_GRID<DTYPE,ATYPE,VTYPE,STYPE>
  (const DTYPE dimension, const ATYPE * axis_size, STYPE * scalar) :
    GRID <DTYPE,ATYPE,VTYPE> (dimension, axis_size)
  // constructor
  {
    this->scalar = scalar;
  }

  template <class DTYPE, class ATYPE, class VTYPE, class STYPE>
  template <class GTYPE> 
  void SCALAR_GRID<DTYPE,ATYPE,VTYPE,STYPE>::Replace
  (const SCALAR_GRID<DTYPE,ATYPE,VTYPE,STYPE> & scalar_grid2, 
   const VTYPE iv0, const VTYPE iv1)
  // replace region from iv0 to iv1 with values from scalar_grid2
  // Precondition: scalar_grid2 has exactly the same dimension
  //   and axis_size as the current scalar grid.
  {
    const DTYPE dimension = this->dimension;
    const ATYPE * axis_size = this->axis_size;
    const STYPE * scalar2 = scalar_grid2.Scalar();
    GTYPE coord0[dimension];
    GTYPE coord1[dimension];
    GTYPE coord2[dimension];
    IJK::PROCEDURE_ERROR error("SCALAR_GRID::Replace");

    if (!Check(scalar_grid2.Dimension(), scalar_grid2.AxisSize(), error))
      { throw error; };

    if (dimension <= 0) { return; }

    compute_coord(iv0, dimension, this->axis_size, coord0);
    compute_coord(iv1, dimension, this->axis_size, coord1);

    // coord2[] represents a point with coordinates of v1 except for coord2[0]
    coord2[0] = coord0[0];
    for (DTYPE d = 1; d < dimension; d++) {
      coord2[d] = coord1[d];
    }

    VTYPE iv2 = get_vertex_index(coord2, dimension, axis_size);
  
    GRID_SIZE_TYPE num_slice_vertices =
      compute_num_grid_vertices(dimension, axis_size, iv0, iv2);

    // vlist contains a region slice orthogonal to the 0 axis
    VTYPE * vlist = new VTYPE[num_slice_vertices];
    get_grid_vertices(dimension, axis_size, iv0, iv2, vlist);

    for (GTYPE c0 = coord0[0]; c0 <= coord1[0]; c0++) {
      for (VTYPE j = 0; j < num_slice_vertices; j++) {
	VTYPE jv = vlist[j];
	scalar[jv] = scalar2[jv];
      }

      // go to next slice
      for (VTYPE j = 0; j < num_slice_vertices; j++) 
	{ vlist[j]++; }
    }

    delete [] vlist;
  }

}

#endif
